package io.github.gretzky.connectivity

import android.Manifest
import android.content.Context
import android.content.pm.PackageManager
import android.net.ConnectivityManager
import android.net.NetworkInfo
import android.net.wifi.ScanResult
import android.net.wifi.WifiInfo
import android.net.wifi.WifiManager
import android.os.Build
import android.telephony.TelephonyManager
import android.telephony.CellSignalStrengthLte
import android.telephony.CellInfoLte
import android.text.format.Formatter

import androidx.core.net.ConnectivityManagerCompat

import com.facebook.react.bridge.ReactApplicationContext
import com.facebook.react.bridge.WritableMap
import com.facebook.react.bridge.WritableNativeMap
import com.facebook.react.bridge.WritableArray
import com.facebook.react.bridge.WritableNativeArray
import com.facebook.react.modules.core.DeviceEventManagerModule

import io.github.gretzky.connectivity.types.CellularGeneration
import io.github.gretzky.connectivity.types.ConnectionType

import javax.annotation.Nonnull
import javax.annotation.Nullable

abstract class RNDUConnectivityReceiver(val reactContext: ReactApplicationContext) {

    val mReactContext: ReactApplicationContext
    val mConnectivityManager: ConnectivityManager
    private val mWifiManager: WifiManager
    private val mTelephonyManager: TelephonyManager

    private var mNoNetworkPermission = false
    private var mConnectionType = ConnectionType.UNKNOWN
    private var mCellularGeneration: CellularGeneration? = null
    private var mIsInternetReachable = false
    private var mIsRoaming: Boolean = false
    private var mCellService: Boolean = false
    private var mCellSignal: Int = 0
    private var mSsid: String = ""
    private var mWifiStrength = 0
    private var mIpAddress = ""
    private var mAvailableNetworks: WritableArray? = null
    private var mIsConnectionExpensive: Boolean = false

    val ssid: String
        get() {
            val rawSSID = mWifiManager.getConnectionInfo().getSSID()
            val formattedSSID = rawSSID.replace("\"", "")
            return formattedSSID
        }

    val wifiStrength: Int
        get() {
            val rssi = mWifiManager.getConnectionInfo().getRssi()
            return mWifiManager.calculateSignalLevel(rssi, 4)
        }

    val ipAddress: String
        get() {
            val ip = mWifiManager.getConnectionInfo().getIpAddress()
            return Formatter.formatIpAddress(ip)
        }

    val isConnectionExpensive: Boolean
        get() {
            return ConnectivityManagerCompat.isActiveNetworkMetered(mConnectivityManager)
        }

    val availableWifiNetworks: WritableArray
        get() {
            val results = mWifiManager.getScanResults()
            val networks = WritableNativeArray()

            for (result in results) {
                val wifiObj = WritableNativeMap()
                if (!result.SSID.isEmpty() && result.level >= -80) {
                    wifiObj.putString("ssid", result.SSID)
                    wifiObj.putInt("strength", mWifiManager.calculateSignalLevel(result.level, 3))
                    networks.pushMap(wifiObj)
                }
            }
            return networks
        }

    val isRoaming: Boolean
        get() {
            if (mReactContext.checkCallingOrSelfPermission(Manifest.permission.READ_PHONE_STATE) === PackageManager.PERMISSION_GRANTED) {
                return mTelephonyManager.getServiceState().getRoaming()
            }
            return false
        }

    val cellService: Boolean
        get() {
            if (mReactContext.checkCallingOrSelfPermission(Manifest.permission.READ_PHONE_STATE) === PackageManager.PERMISSION_GRANTED) {
                val state = mTelephonyManager.getServiceState().getState()
                return if (state == 0) {
                    true
                } else false
            }
            return false
        }

    val cellSignal: Int
        get() {
            if (mReactContext.checkCallingOrSelfPermission(Manifest.permission.ACCESS_COARSE_LOCATION) === PackageManager.PERMISSION_GRANTED) {
                val cellInfo = mTelephonyManager.getAllCellInfo().get(0) as CellInfoLte
                return cellInfo.getCellSignalStrength().getLevel()
            }
            return 0
        }

    init {
        mReactContext = reactContext as ReactApplicationContext
        mConnectivityManager = reactContext.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
        mWifiManager = reactContext.getSystemService(Context.WIFI_SERVICE) as WifiManager
        mTelephonyManager = reactContext.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager

        mCellSignal = cellSignal
        mCellService = cellService
        mIsRoaming = isRoaming
        mSsid = ssid
        mIpAddress = ipAddress
        mWifiStrength = wifiStrength
        mAvailableNetworks = availableWifiNetworks
        mIsConnectionExpensive = isConnectionExpensive
    }

    internal abstract fun register()

    internal abstract fun unregister()

    fun sendChangeEvent() {
        reactContext
            .getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter::class.java)
            .emit("connectivity", createEventMap())
    }

    fun updateConnectivity(
        @Nonnull connectionType: ConnectionType,
        isInternetReachable: Boolean,
        isConnectionExpensive: Boolean
    ) {
        // android will throw multiple broadcasts for the same connection, so we ensure that we only
        // emit an event when there's an actual change
        val connectionTypeChanged = connectionType != mConnectionType
        val isInternetReachableChanged = isInternetReachable != mIsInternetReachable
        val isConnectionExpensiveChanged = isConnectionExpensive != mIsConnectionExpensive

        when {
            connectionTypeChanged -> {
                mConnectionType = connectionType
                sendChangeEvent()
            }
            isInternetReachableChanged -> {
                mIsInternetReachable = isInternetReachable
                sendChangeEvent()
            }
            isConnectionExpensiveChanged -> {
                mIsConnectionExpensive = isConnectionExpensive
                sendChangeEvent()
            }
            else -> {}
        }
    }

    fun updateCellularAttributes(@Nullable cellularGeneration: CellularGeneration, isRoaming: Boolean, cellService: Boolean, cellSignal: Int) {
        val cellularGenerationChanged = cellularGeneration != mCellularGeneration
        val roamingChanged = isRoaming != mIsRoaming
        val serviceChanged = cellService != mCellService
        val signalChanged = cellSignal != mCellSignal

        when {
            cellularGenerationChanged -> {
                mCellularGeneration = cellularGeneration
                sendChangeEvent()
            }
            roamingChanged -> {
                mIsRoaming = isRoaming
                sendChangeEvent()
            }
            serviceChanged -> {
                mCellService = cellService
                sendChangeEvent()
            }
            signalChanged -> {
                mCellSignal = cellSignal
                sendChangeEvent()
            }
            else -> {}
        }
    }

    fun updateWifiAttributes(ssid: String, ip: String, strength: Int) {
        val ssidChanged = ssid != mSsid
        val ipChanged = ip != mIpAddress
        val strengthChanged = strength != mWifiStrength

        when {
            ssidChanged -> {
                mSsid = ssid
                sendChangeEvent()
            }
            ipChanged -> {
                mIpAddress = ip
                sendChangeEvent()
            }
            strengthChanged -> {
                mWifiStrength = strength
                sendChangeEvent()
            }
            else -> {}
        }
    }

    private fun formatWifiSignalStrength(strength: Int): Int {
        // ranges from 1-4 with 1 being crap and 4 being great
        return when {
            strength >= -50 -> 4
            strength < -50 && strength >= -60 -> 3
            strength < -60 && strength >= -70 -> 2
            strength < -70 && strength >= -80 -> 1
            else -> 0 //unusable if strength isn't a valid int, or of its lower than -80dBm
        }
    }

    fun createEventMap(): WritableMap {
        var event: WritableMap = WritableNativeMap()
        var connectivity: WritableMap = WritableNativeMap()

        val wifiEnabled = mWifiManager.isWifiEnabled()
        val isConnected = !mConnectionType.equals(ConnectionType.NONE) && !mConnectionType.equals(ConnectionType.UNKNOWN)

        connectivity.putString("type", mConnectionType.label)
        connectivity.putBoolean("isConnected", isConnected)

        val details: WritableMap = WritableNativeMap()
        if (isConnected) {
            details.putBoolean("isConnectionExpensive", mIsConnectionExpensive)

            details.putBoolean("isInternetReachable", mIsInternetReachable)

            if (mConnectionType.equals(ConnectionType.CELLULAR)) {
                details.putString("cellularGeneration", mCellularGeneration!!.label)
                details.putString("carrier", mTelephonyManager.getNetworkOperatorName())
                details.putBoolean("isInService", mCellService)
                details.putBoolean("isRoaming", mIsRoaming)
                details.putInt("strength", mCellSignal)
            }

            if (mConnectionType.equals(ConnectionType.WIFI)) {
                details.putString("ssid", mSsid)

                details.putInt("strength", mWifiStrength)

                details.putString("ip", mIpAddress)
            }
        } else if (!isConnected && wifiEnabled) {
            // if we have no connection but wifi is enabled, get an array of nearby networks to connect to
            details.putArray("availableNetworks", mAvailableNetworks)
        }

        connectivity.putMap("details", details)
        event.putMap("connectivity", connectivity)

        return event
    }
}
