import * as React from "react";
import Icon from "react-native-vector-icons/MaterialCommunityIcons";
import { ConnectivityComponentProps } from "../util/types";

const renderIcon = (name: string, color?: string) => <Icon size={20} color={color} name={name} />;

/**
 * ConnectivityIcon - render an icon that displays connectivity strength (wifi or cell)
 */
const ConnectivityIcon: React.FC<ConnectivityComponentProps> = ({ state, color }) => {
  const strength = state.connectivity.details.strength;
  const isConnected = state.connectivity.isConnected;
  const isReachable = state.connectivity.details.isInternetReachable;
  const connectionType = state.connectivity.type;

  const isWifi = connectionType === "wifi";
  const isCell = connectionType === "cellular";

  /**
   * handleIconColor - throw 'red' if the icon name is less than 1 (low strength) or a string (strength unavailable)
   * @param name the name of the icon to use
   */
  const handleIconColor = (name: number | string): string =>
    name <= 1 || typeof name === "string" ? "#ee4444" : color;

  /**
   * renderWifiIcon - render a wifi connectivity icon with the given strength
   * @param name the name of the icon to use (corresponds to the strength)
   */
  const renderWifiIcon = (name: number | string) =>
    renderIcon(`wifi-strength-${name}${!isReachable ? "-alert" : ""}`, handleIconColor(name)); // if the internet isn't reachable, but available, use an alert icon

  /**
   * renderCellularIcon - render a cellular connectivity icon with the given strength
   * @param name the name of the icon to use (corresponds to the strength)
   */
  const renderCellularIcon = (name: number | string) => renderIcon(`signal-cellular-${name}`, handleIconColor(name));

  /**
   * getAppropriateIcon - gets the icon for the given connectivity type (wifi or cellular)
   * @param name the name of the icon to use (corresponds to the strength)
   */
  const getAppropriateIcon = (name: number | string) =>
    isWifi ? renderWifiIcon(name) : isCell ? renderCellularIcon(name) : null;

  // show outlined (or alert, if available) icon if we aren't connected
  if (!isConnected) {
    return isWifi ? renderWifiIcon("alert-outline") : renderCellularIcon("outline");
  }

  switch (strength) {
    case 4:
      // the cellular icon only has 3 bars (but the strength still goes up to 4)
      // we throw 'outline' in case we catch an exception here
      return getAppropriateIcon(isWifi ? 4 : isCell ? 3 : "outline");
    case 3:
      return getAppropriateIcon(3);
    case 2:
      return getAppropriateIcon(2);
    case 1:
      return getAppropriateIcon(1);
    default:
      return getAppropriateIcon("alert-outline");
  }
};

ConnectivityIcon.defaultProps = {
  color: "black"
};

export default ConnectivityIcon;
